# This file is part of colorspacious
# Copyright (C) 2014 Nathaniel Smith <njs@pobox.com>
# See file LICENSE.txt for license information.

from __future__ import division

from collections import namedtuple

import numpy as np

from .illuminants import as_XYZ100_w

__all__ = [
    "CIECAM02Surround", "CIECAM02Space", "NegativeAError",
    "JChQMsH",
]


# F, c, Nc: surround parameters
#            F     c      Nc
# Average   1.0   0.69   1.0
# Dim       0.9   0.59   0.9
# Dark      0.8   0.525  0.8
# See: https://github.com/njsmith/colorspacious/issues/14
CIECAM02Surround = namedtuple("CIECAM02Surround", ["F", "c", "N_c"])
CIECAM02Surround.AVERAGE = CIECAM02Surround(1.0, 0.69,  1.0)
CIECAM02Surround.DIM     = CIECAM02Surround(0.9, 0.59,  0.9)
CIECAM02Surround.DARK    = CIECAM02Surround(0.8, 0.525, 0.8)

JChQMsH = namedtuple("JChQMsH", ["J", "C", "h", "Q", "M", "s", "H"])

M_CAT02 = np.asarray([[ 0.7328,  0.4296, -0.1624],
                      [-0.7036,  1.6975,  0.0061],
                      [ 0.0030,  0.0136,  0.9834]])

M_HPE = np.asarray([[ 0.38971,  0.68898, -0.07868],
                    [-0.22981,  1.18340,  0.04641],
                    [ 0.00000,  0.00000,  1.00000]])

# These are very well-conditioned matrices (condition numbers <4), so just
# taking the inverse is fine, and it simplifies things below.
M_CAT02_inv = np.linalg.inv(M_CAT02)
M_HPE_M_CAT02_inv = np.dot(M_HPE, M_CAT02_inv)
M_CAT02_M_HPE_inv = np.dot(M_CAT02, np.linalg.inv(M_HPE))

h_i = np.asarray([20.14,  90.00, 164.25, 237.53, 380.14])
e_i = np.asarray([ 0.8,    0.7,    1.0,    1.2,    0.8])
H_i = np.asarray([ 0.0,  100.0,  200.0,  300.0,  400.0])

def broadcasting_matvec(A, B):
    # We want to handle two cases that come up a bunch below.
    # B is always a vector, or collection of vectors. So it has shape
    #    (..., j)
    # where j is the number of entries in each vector.
    # A is either a matrix or a vector, and we want to broadcast np.dot(A,
    # B_vec) over all possible B_vecs.
    # When A has shape (i, j), this means our result should have shape
    #    (..., i)
    # and when A has shape (j,), this means our result should have shape
    #    (...)
    # So a generalization is that given
    #   A.shape == (...1, j)
    #   B.shape == (...2, j)
    # we want a result with shape
    #   (...2, ...1)
    # it turns out that this is the magic incantation for doing that:
    return np.inner(B, A)

def require_exactly_one(**kwargs):
    non_None = sum(v is not None for v in kwargs.values())
    if non_None != 1:
        raise ValueError("Exactly one of %s must be specified"
                         % (", ".join(kwargs)))

class NegativeAError(ValueError):
    """A :class:`ValueError` that can be raised when converting to CIECAM02.

    See :meth:`CIECAM02Space.XYZ100_to_CIECAM02` for details.
    """
    pass

class CIECAM02Space(object):
    """An object representing a particular set of CIECAM02 viewing conditions.

    :param XYZ100_w: The whitepoint. Either a string naming one of the known
         standard whitepoints like ``"D65"``, or else a point in XYZ100 space.
    :param Y_b: Background luminance.
    :param L_A: Luminance of the adapting field (in cd/m^2).
    :param surround: A :class:`CIECAM02Surround` object.

    """
    def __init__(self, XYZ100_w, Y_b, L_A,
                 surround=CIECAM02Surround.AVERAGE):
        self.XYZ100_w = as_XYZ100_w(XYZ100_w)
        # as_XYZ100_w allows for multiple whitepoints to be returned, but we
        # aren't vectorized WRT whitepoint
        if self.XYZ100_w.shape != (3,):
            raise ValueError("Hey! XYZ100_w should have shape (3,)!")
        self.Y_b = float(Y_b)
        self.L_A = float(L_A)
        self.surround = surround
        self.F = float(surround.F)
        self.c = float(surround.c)
        self.N_c = float(surround.N_c)

        self.RGB_w = np.dot(M_CAT02, self.XYZ100_w)
        self.D = self.F * (1 - (1/3.6) * np.exp((-self.L_A - 42) / 92))
        self.D = np.clip(self.D, 0, 1)

        self.D_RGB = self.D * self.XYZ100_w[1] / self.RGB_w + 1 - self.D
        # Fairchild (2013), pages 290-292, recommends using this equation
        # instead, though notes that it doesn't make much difference as part
        # of a full CIECAM02 system. (It matters more if you're only using
        # pieces.)
        #self.D_RGB = self.D * 100 / self.RGB_w + 1 - self.D
        self.k = 1 / (5 * self.L_A + 1)
        self.F_L = (0.2 * self.k ** 4 * (5 * self.L_A)
                    + 0.1 * (1 - self.k**4)**2 * (5 * self.L_A) ** (1./3))
        self.n = self.Y_b / self.XYZ100_w[1]
        self.z = 1.48 + np.sqrt(self.n)
        self.N_bb = 0.725 * (1 / self.n)**0.2
        self.N_cb = self.N_bb  #??

        self.RGB_wc = self.D_RGB * self.RGB_w
        self.RGBprime_w = np.dot(M_HPE_M_CAT02_inv, self.RGB_wc)
        tmp = ((self.F_L * self.RGBprime_w) / 100) ** 0.42
        self.RGBprime_aw = 400 * (tmp / (tmp + 27.13)) + 0.1
        self.A_w = ((np.dot([2, 1, 1. / 20], self.RGBprime_aw) - 0.305)
                    * self.N_bb)

    def __repr__(self):
        surround_string = ", surround=%r" % (self.surround,)
        if self.surround == CIECAM02Surround.AVERAGE:
            surround_string = ""
        return "%s(%r, %r, %r%s) " % (
            self.__class__.__name__,
            list(self.XYZ100_w),
            self.Y_b,
            self.L_A,
            surround_string)

    # XYZ100 must have shape (3,) or (3, n)
    def XYZ100_to_CIECAM02(self, XYZ100, on_negative_A="raise"):
        """Computes CIECAM02 appearance correlates for the given tristimulus
        value(s) XYZ (normalized to be on the 0-100 scale).

        Example: ``vc.XYZ100_to_CIECAM02([30.0, 45.5, 21.0])``

        :param XYZ100: An array-like of tristimulus values. These should be
          given on the 0-100 scale, not the 0-1 scale. The array-like should
          have shape ``(..., 3)``; e.g., you can use a simple 3-item list
          (shape = ``(3,)``), or to efficiently perform multiple computations
          at once, you could pass a higher-dimensional array, e.g. an image.
        :arg on_negative_A: A known infelicity of the CIECAM02 model is that
          for some inputs, the achromatic signal :math:`A` can be negative,
          which makes it impossible to compute :math:`J`, :math:`C`,
          :math:`Q`, :math:`M`, or :math:`s` -- only :math:`h`: and :math:`H`
          are spared. (See, e.g., section 2.6.4.1 of :cite:`Luo-CIECAM02` for
          discussion.) This argument allows you to specify a strategy for
          handling such points. Options are:

          * ``"raise"``: throws a :class:`NegativeAError` (a subclass of
            :class:`ValueError`)
          * ``"nan"``: return not-a-number values for the affected
            elements. (This may be particularly useful if converting a large
            number of points at once.)

        :returns: A named tuple of type :class:`JChQMsH`, with attributes
          ``J``, ``C``, ``h``, ``Q``, ``M``, ``s``, and ``H`` containing the
          CIECAM02 appearance correlates.

        """

        #### Argument checking

        XYZ100 = np.asarray(XYZ100, dtype=float)
        if XYZ100.shape[-1] != 3:
            raise ValueError("XYZ100 shape must be (..., 3)")

        #### Step 1

        RGB = broadcasting_matvec(M_CAT02, XYZ100)

        #### Step 2

        RGB_C = self.D_RGB * RGB

        #### Step 3

        RGBprime = broadcasting_matvec(M_HPE_M_CAT02_inv, RGB_C)

        #### Step 4

        RGBprime_signs = np.sign(RGBprime)

        tmp = (self.F_L * RGBprime_signs * RGBprime / 100) ** 0.42
        RGBprime_a = RGBprime_signs * 400 * (tmp / (tmp + 27.13)) + 0.1

        #### Step 5

        a = broadcasting_matvec([1, -12. / 11, 1. / 11], RGBprime_a)
        b = broadcasting_matvec([1. / 9, 1. / 9, -2. / 9], RGBprime_a)
        h_rad = np.arctan2(b, a)
        h = np.rad2deg(h_rad) % 360

        # #### Step 6

        # hprime = h, unless h < 20.14, in which case hprime = h + 360.
        hprime = np.select([h < h_i[0], True], [h + 360, h])
        # we use 0-based indexing, so our i is one less than the reference
        # formulas' i.
        i = np.searchsorted(h_i, hprime, side="right") - 1
        tmp = (hprime - h_i[i]) / e_i[i]
        H = H_i[i] + ((100 * tmp)
                      / (tmp + (h_i[i + 1] - hprime) / e_i[i + 1]))

        #### Step 7

        A = ((broadcasting_matvec([2, 1, 1. / 20], RGBprime_a) - 0.305)
             * self.N_bb)

        if on_negative_A == "raise":
            if np.any(A < 0):
                raise NegativeAError("attempted to convert a tristimulus "
                                     "value whose achromatic signal was "
                                     "negative, and on_negative_A=\"raise\"")
        elif on_negative_A == "nan":
            A = np.select([A < 0, True], [np.nan, A])
        else:
            raise ValueError("Invalid on_negative_A argument: got %r, "
                             "expected \"raise\" or \"nan\""
                             % (on_negative_A,))

        #### Step 8

        J = 100 * (A / self.A_w) ** (self.c * self.z)

        #### Step 9

        Q = self._J_to_Q(J)

        #### Step 10

        e = (12500. / 13) * self.N_c * self.N_cb * (np.cos(h_rad + 2) + 3.8)
        t = (e * np.sqrt(a ** 2 + b ** 2)
             / broadcasting_matvec([1, 1, 21. / 20], RGBprime_a))

        C = t**0.9 * (J / 100)**0.5 * (1.64 - 0.29**self.n)**0.73
        M = C * self.F_L**0.25
        s = 100 * (M / Q)**0.5

        return JChQMsH(J, C, h, Q, M, s, H)

    def _J_to_Q(self, J):
        return ((4 / self.c) * (J / 100) ** 0.5
                * (self.A_w + 4) * self.F_L**0.25)

    def CIECAM02_to_XYZ100(self, J=None, C=None, h=None,
                           Q=None, M=None, s=None, H=None):
        """Return the unique tristimulus values that have the given CIECAM02
        appearance correlates under these viewing conditions.

        You must specify 3 arguments:

        * Exactly one of ``J`` and ``Q``
        * Exactly one of ``C``, ``M``, and ``s``
        * Exactly one of ``h`` and ``H``.

        Arguments can be vectors, in which case they will be broadcast against
        each other.

        Returned tristimulus values will be on the 0-100 scale, not the 0-1
        scale.
        """

        #### Argument checking

        require_exactly_one(J=J, Q=Q)
        require_exactly_one(C=C, M=M, s=s)
        require_exactly_one(h=h, H=H)

        if J is not None:
            J = np.asarray(J, dtype=float)
        if C is not None:
            C = np.asarray(C, dtype=float)
        if h is not None:
            h = np.asarray(h, dtype=float)
        if Q is not None:
            Q = np.asarray(Q, dtype=float)
        if M is not None:
            M = np.asarray(M, dtype=float)
        if s is not None:
            s = np.asarray(s, dtype=float)
        if H is not None:
            H = np.asarray(H, dtype=float)

        #### Step 1: conversions to get JCh

        if J is None:
            J = 6.25 * ((self.c * Q) / ((self.A_w + 4) * self.F_L**0.25)) ** 2

        if C is None:
            if M is not None:
                C = M / self.F_L**0.25
            else:
                assert s is not None
                # when starting from s, we need Q
                if Q is None:
                    Q = self._J_to_Q(J)
                C = (s / 100) ** 2 * (Q / self.F_L**0.25)

        if h is None:
            i = np.searchsorted(H_i, H, side="right") - 1
            # BROKEN:
            num1 = (H - H_i[i]) * (e_i[i + 1] * h_i[i] - e_i[i] * h_i[i + 1])
            num2 = -100 * h_i[i] * e_i[i + 1]
            denom1 = (H - H_i[i]) * (e_i[i + 1] - e_i[i])
            denom2 = -100 * e_i[i + 1]
            hprime = (num1 + num2) / (denom1 + denom2)
            h = np.select([hprime > 360, True], [hprime - 360, hprime])

        J, C, h = np.broadcast_arrays(J, C, h)
        target_shape = J.shape

        # 0d arrays break indexing stuff
        if J.ndim == 0:
            J = np.atleast_1d(J)
            C = np.atleast_1d(C)
            h = np.atleast_1d(h)

        #### Step 2

        t = (C
             / (np.sqrt(J / 100) * (1.64 - 0.29**self.n) ** 0.73)
            ) ** (1 / 0.9)
        e_t = 0.25 * (np.cos(np.deg2rad(h) + 2) + 3.8)
        A = self.A_w * (J / 100) ** (1 / (self.c * self.z))

        # an awkward way of calculating 1/t such that 1/0 -> inf
        with np.errstate(divide="ignore"):
            one_over_t = 1 / t
        one_over_t = np.select([np.isnan(one_over_t), True],
                               [np.inf, one_over_t])

        p_1 = (50000. / 13) * self.N_c * self.N_cb * e_t * one_over_t
        p_2 = A / self.N_bb + 0.305
        p_3 = 21. / 20

        #### Step 3

        sin_h = np.sin(np.deg2rad(h))
        cos_h = np.cos(np.deg2rad(h))

        # to avoid divide-by-zero (or divide-by-eps) issues, we use different
        # computations when |sin_h| > |cos_h| and vice-versa
        num = p_2 * (2 + p_3) * (460. / 1403)
        denom_part2 = (2 + p_3) * (220. / 1403)
        denom_part3 = (-27. / 1403) + p_3 * (6300. / 1403)

        a = np.empty_like(h)
        b = np.empty_like(h)

        small_cos = (np.abs(sin_h) >= np.abs(cos_h))

        # NB denom_part2 and denom_part3 are scalars
        b[small_cos] = (num[small_cos]
                        / (p_1[small_cos] / sin_h[small_cos]
                           + (denom_part2
                              * cos_h[small_cos] / sin_h[small_cos])
                           + denom_part3))
        a[small_cos] = b[small_cos] * cos_h[small_cos] / sin_h[small_cos]

        a[~small_cos] = (num[~small_cos]
                         / (p_1[~small_cos] / cos_h[~small_cos]
                            + denom_part2
                            + (denom_part3
                               * sin_h[~small_cos] / cos_h[~small_cos])))
        b[~small_cos] = a[~small_cos] * sin_h[~small_cos] / cos_h[~small_cos]

        #### Step 4

        p2ab = np.concatenate((p_2[..., np.newaxis],
                               a[..., np.newaxis],
                               b[..., np.newaxis]),
                              axis=-1)
        RGBprime_a_matrix = (1. / 1403
                             * np.asarray([[ 460,  451,   288],
                                           [ 460, -891,  -261],
                                           [ 460, -220, -6300]], dtype=float))

        RGBprime_a = broadcasting_matvec(RGBprime_a_matrix, p2ab)

        #### Step 5

        RGBprime = (np.sign(RGBprime_a - 0.1)
                    * (100 / self.F_L)
                    * ((27.13 * np.abs(RGBprime_a - 0.1))
                       / (400 - np.abs(RGBprime_a - 0.1))) ** (1 / 0.42))

        #### Step 6

        RGB_C = broadcasting_matvec(M_CAT02_M_HPE_inv, RGBprime)

        #### Step 7

        RGB = RGB_C / self.D_RGB

        #### Step 8

        XYZ100 = broadcasting_matvec(M_CAT02_inv, RGB)

        XYZ100 = XYZ100.reshape(target_shape + (3,))

        return XYZ100

CIECAM02Space.sRGB = CIECAM02Space(
    # sRGB specifies a D65 monitor and a D50 ambient. CIECAM02 doesn't really
    # know how to deal with this discrepancy; it appears that the usual thing
    # to do is just to use D65 for the whitepoint.
    XYZ100_w="D65",
    Y_b=20,
    # To compute L_A:
    #   illuminance in lux / pi = luminance in cd/m^2
    #   luminance in cd/m^2 / 5 = L_A (the "grey world assumption")
    # See Moroney (2000), "Usage guidelines for CIECAM97s".
    # sRGB illuminance is 64 lux.
    L_A=(64 / np.pi) / 5,
    surround=CIECAM02Surround.AVERAGE)


################################################################
# Tests
################################################################

def check_roundtrip(vc, XYZ100):
    try:
        values = vc.XYZ100_to_CIECAM02(XYZ100, on_negative_A="raise")
    except NegativeAError:
        # don't expect to be able to round-trip these values
        return
    for kwarg1 in ["J", "Q"]:
        for kwarg2 in ["C", "M", "s"]:
            for kwarg3 in ["h", "H"]:
                got = vc.CIECAM02_to_XYZ100(**{kwarg1: getattr(values, kwarg1),
                                            kwarg2: getattr(values, kwarg2),
                                            kwarg3: getattr(values, kwarg3)})
                assert np.allclose(got, XYZ100)

def test_gold():
    from .gold_values import XYZ100_CIECAM02_gold

    for t in XYZ100_CIECAM02_gold:
        got = t.vc.XYZ100_to_CIECAM02(t.XYZ100)
        for i in range(len(got)):
            assert np.allclose(got[i], t.expected[i], atol=1e-05)
        check_roundtrip(t.vc, t.XYZ100)

def test_inverse():
    r = np.random.RandomState(0)
    XYZ100_values = [[0, 0, 0], [50, 50, 50]]
    for i in range(10):
        XYZ100_values.append(r.uniform(high=100, size=3))

    for illuminant in ["C", "D50", "D65"]:
        for Y_b in [20, 18]:
            for L_A in [30, 300]:
                for surround in [CIECAM02Surround.AVERAGE,
                                 CIECAM02Surround.DIM,
                                 CIECAM02Surround.DARK]:
                    vc = CIECAM02Space(illuminant, Y_b, L_A, surround)
                    for XYZ100 in XYZ100_values:
                        check_roundtrip(vc, XYZ100)

def test_misc():
    from nose.tools import assert_raises
    # Only one whitepoint can be specified
    assert_raises(ValueError, CIECAM02Space,
                  [[20, 100, 80], [80, 100, 20]], 20, 30)

    # smoke test
    repr(CIECAM02Space.sRGB)
    repr(CIECAM02Space("D65", 20, 4, surround=CIECAM02Surround.DIM))

    # input shape check
    assert_raises(ValueError,
                  CIECAM02Space.sRGB.XYZ100_to_CIECAM02,
                  np.ones((10, 4)))

    # on_negative_A validity check
    assert_raises(ValueError,
                  CIECAM02Space.sRGB.XYZ100_to_CIECAM02,
                  np.ones((10, 3)),
                  on_negative_A="asdfasdf")

def test_exactly_one():
    from nose.tools import assert_raises
    vc = CIECAM02Space.sRGB

    # Redundant specifications not allowed
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, J=1, C=1, h=1, Q=1)
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, J=1, C=1, h=1, M=1)
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, J=1, C=1, h=1, s=1)
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, J=1, C=1, h=1, H=1)

    # Underspecified colors not allowed either
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, J=1, C=1)
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, J=1, h=1)
    assert_raises(ValueError, vc.CIECAM02_to_XYZ100, C=1, h=1)

def test_vectorized():
    vc = CIECAM02Space.sRGB

    XYZ100s = [[20, 30, 40], [40, 30, 20]]
    CIECAM02s = vc.XYZ100_to_CIECAM02(XYZ100s)

    for i, XYZ100 in enumerate(XYZ100s):
        CIECAM02 = vc.XYZ100_to_CIECAM02(XYZ100)
        for j in range(len(CIECAM02)):
            assert np.allclose(CIECAM02[j], CIECAM02s[j][i])

    check_roundtrip(vc, XYZ100s)

def test_on_negative_A():
    from nose.tools import assert_raises

    vc = CIECAM02Space("D65", 20, 30)
    bad_XYZ100 = [8.71292997, 2.02183974, 83.26198455]
    good_XYZ100 = [20, 30, 40]

    assert_raises(NegativeAError, vc.XYZ100_to_CIECAM02, bad_XYZ100)
    assert_raises(NegativeAError, vc.XYZ100_to_CIECAM02,
                  [bad_XYZ100, good_XYZ100])
    assert_raises(NegativeAError, vc.XYZ100_to_CIECAM02, bad_XYZ100,
                  on_negative_A="raise")
    assert_raises(NegativeAError, vc.XYZ100_to_CIECAM02,
                  [bad_XYZ100, good_XYZ100],
                  on_negative_A="raise")

    bad_CIECAM02 = vc.XYZ100_to_CIECAM02(bad_XYZ100, on_negative_A="nan")
    for bad_attr in "JCQMs":
        assert np.isnan(getattr(bad_CIECAM02, bad_attr))
    assert np.allclose(bad_CIECAM02.h, 205.80008)
    assert np.allclose(bad_CIECAM02.H, 261.11054)

    mixed_CIECAM02 = vc.XYZ100_to_CIECAM02([bad_XYZ100, good_XYZ100],
                                           on_negative_A="nan")
    for bad_attr in "JCQMs":
        assert np.all(np.isnan(getattr(mixed_CIECAM02, bad_attr))
                      == [True, False])
    assert np.all(np.isnan(mixed_CIECAM02.h) == [False, False])
    assert np.all(np.isnan(mixed_CIECAM02.H) == [False, False])



# This file is part of colorspacious
# Copyright (C) 2015 Nathaniel Smith <njs@pobox.com>
# See file LICENSE.txt for license information.

import numpy as np

from collections import namedtuple, defaultdict

__all__ = ["Edge", "MATCH", "ANY", "TransformGraph"]

################################################################
# Basic types and sentinels
################################################################

# We give these a string id for ease of debugging, but they are really
# singleton sentinels compared via is/is not.
class Placeholder(object):
    def __init__(self, id):
        self.id = id

    def __repr__(self):  # pragma: no cover
        return "<{0}>".format(self.id)

MATCH = Placeholder("MATCH")
ANY = Placeholder("ANY")

START = Placeholder("START")
END = Placeholder("END")

Edge = namedtuple("Edge", ["start", "end", "transform"])
Path = namedtuple("Path", ["nodes", "transforms"])

################################################################
# Invariant checking
################################################################

def assert_(b):
    if not b:
        raise AssertionError

def check_node(node, allowed_placeholders):
    assert_(isinstance(node, dict))
    assert_(isinstance(node.get("name"), str))
    for v in node.values():
        if isinstance(v, Placeholder):
            assert_(v in allowed_placeholders)

def test_check_node():
    from nose.tools import assert_raises
    assert_raises(AssertionError, check_node, {}, [])
    check_node({"name": "asdf"}, [])
    assert_raises(AssertionError, check_node, {"name": 1}, [])
    check_node({"name": "asdf", "a": "b"}, [])
    assert_raises(AssertionError, check_node,
                  {"name": "asdf", "a": MATCH}, [ANY])
    check_node({"name": "asdf", "a": ANY, "b": "c"}, [ANY])

# edge_node = edge node, can have MATCH or ANY or concrete values
# used in primitive edges
def check_edge_node(edge_node):
    check_node(edge_node, set([MATCH, ANY]))

# path_node = path node, can have START or END or concrete values
# used in precomputed paths
def check_path_node(path_node):
    check_node(path_node, set([START, END]))

# concrete_node = concrete node, can only have concrete values
# used in queries
def check_concrete_node(concrete_node):
    check_node(concrete_node, set())

def check_edge(edge):
    check_edge_node(edge.start)
    check_edge_node(edge.end)
    NOTHING = object()
    for k in set(edge.start).union(edge.end):
        start_v = edge.start.get(k, NOTHING)
        end_v = edge.end.get(k, NOTHING)
        # Every _MATCH must have a matching _MATCH.
        if MATCH in (start_v, end_v):
            assert_(start_v is end_v is MATCH)
        # For simplicity, ANY must not match anything.
        # In principle you could have a transform
        #    {"foo": "concrete_value"} -> {"foo": ANY}
        # or even
        #    {"foo": ANY} -> {"foo": ANY}
        # where the transform can turn any setting of foo into any other. But
        # we currently don't need this, it complicates things, and makes it
        # harder to pass the values to the transform, so YAGNI.
        if ANY in (start_v, end_v):
            assert_(set([start_v, end_v]) == set([ANY, NOTHING]))

def test_check_edge():
    from nose.tools import assert_raises

    check_edge(Edge({"name": "foo"}, {"name": "bar"}, "fake transform"))
    check_edge(Edge({"name": "foo", "start_any": ANY, "match": MATCH},
                    {"name": "bar", "end_any": ANY, "match": MATCH},
                    "fake transform"))
    for bad in [
            Edge({"name": 1}, {"name": "bar"}, "fake transform"),
            Edge({"name": "foo"}, {"name": 1}, "fake transform"),
            Edge({"name": "foo"}, {"name": "bar", "other": START},
                 "fake transform"),
            Edge({"name": "foo", "a": ANY}, {"name": "bar", "a": ANY},
                 "fake transform"),
            Edge({"name": "foo", "a": MATCH}, {"name": "bar", "a": "asdf"},
                 "fake transform"),
            Edge({"name": "foo", "a": "asdf"}, {"name": "bar", "a": MATCH},
                 "fake transform"),
            ]:
        assert_raises(AssertionError, check_edge, bad)

def check_path(path):
    for node in path.nodes:
        check_path_node(node)
    for start, end in zip(path.nodes[:-1], path.nodes[1:]):
        for k in set(start).union(end):
            if k != "name" and k in start and k in end:
                assert_(start[k] == end[k])
            # START and END must have an unbroken chain from the beginning or
            # end of the path respectively, b/c the only way they can legally
            # appear in the middle is via MATCH constraints.
            if start.get(k) is END:
                assert_(end.get(k) is END)
            if end.get(k) is START:
                assert_(start.get(k) is START)
    assert_(len(path.nodes) == len(path.transforms) + 1)

def test_check_path():
    from nose.tools import assert_raises

    check_path(Path(({"name": "foo"}, {"name": "bar"}), ("t1",)))
    check_path(Path(({"name": "foo"},
                     {"name": "bar"},
                     {"name": "baz"}),
                    ("t1", "t2")))
    for bad in [
            # must be path nodes
            Path([{"name": 1}, {"name": "bar"}], ["t1"]),
            Path([{"name": "foo"}, {"name": 1}], ["t1"]),
            Path([{"name": "foo", "a": ANY}, {"name": "bar"}], ["t1"]),
            Path([{"name": "foo"}, {"name": "bar", "b": MATCH}], ["t1"]),
            # no changing values on a transition
            Path([{"name": "foo", "a": 1}, {"name": "bar", "a": 2}], ["t1"]),
            # no dropping END
            Path([{"name": "foo", "a": END}, {"name": "bar"}], ["t1"]),
            Path([{"name": "foo", "a": END}, {"name": "bar", "a": 2}], ["t1"]),
            # no spontaneous generation of START
            Path([{"name": "foo"}, {"name": "bar", "a": START}], ["t1"]),
            Path([{"name": "foo", "a": 1}, {"name": "bar", "a": START}], ["t1"]),
            # length mismatches
            Path([{"name": "foo"}, {"name": "bar"}], []),
            Path([{"name": "foo"}, {"name": "bar"}], ["t1", "t2"]),
            ]:
        assert_raises(AssertionError, check_path, bad)

################################################################
# Dict manipulation utilities
################################################################

# Utilities for replacing values in a dict in-place
def _replace_values(d, replacements):
    # replacements is {old_value: new_value}
    for k, v in d.items():
        if d[k] in replacements:
            d[k] = replacements[v]

def test__replace_values():
    d = {"a":  1, "b": 2, "c":  1}
    _replace_values(d, {1: 11})
    assert_(d == {"a": 11, "b": 2, "c": 11})

def _fill_values_from(d, marker, template):
    # replaces all instances of 'marker' in d with corresponding value from
    # 'template'
    for k, v in d.items():
        if d[k] is marker:
            d[k] = template[k]

def test__fill_values_from():
    d = {"a": 1, "b": 2, "c": 1}
    _fill_values_from(d, 1, {"a": "a1", "b": "b1", "c": "c1"})
    assert_(d == {"a": "a1", "b": 2, "c": "c1"})

################################################################
# Graph calculations
################################################################

def trivial_path(edge_node):
    path_node = dict(edge_node)
    _replace_values(path_node, {MATCH: START, ANY: START})
    return Path([path_node], [])

def test_trivial_path():
    assert_(trivial_path({"name": "start",
                          "foo": MATCH,
                          "bar": ANY,
                          "baz": 1})
            == Path([{"name": "start",
                      "foo": START,
                      "bar": START,
                      "baz": 1}],
                    []))

def try_extend_path(path, edge):
    # Returns a new Path, or None if this edge can't extend this path.

    # First, check that the last node of the path is compatible with the first
    # node of the edge. This might require imposing new constraints on the
    # path, e.g. if our path is
    #
    #   {"name": "a", "x": START} -> {"name": "b", "x": START}
    #
    # and we want to extend it with an edge
    #
    #   {"name": "b", "x": 0} -> {"name": "c"}
    #
    # then that should produce the path
    #
    #   {"name": "a", "x": 0} -> {"name": "b", "x": 0} -> {"name": "c"}
    #
    # where we match the overlapping nodes and then propagated the new
    # constraint back.
    if path.nodes[-1].keys() != edge.start.keys():
        return None
    path_fill_ins = defaultdict(dict)
    for k in path.nodes[-1]:
        path_value = path.nodes[-1][k]
        edge_value = edge.start[k]
        if path_value is START:
            # START matches anything, but if the edge has a concrete value
            # then we need to add that constraint
            if not isinstance(edge_value, Placeholder):
                path_fill_ins[START][k] = edge_value
        elif path_value is END:
            # END matches concrete values and MATCH, but not ANY (because if
            # it matched ANY, then the END chain would be lost)
            if isinstance(edge_value, Placeholder):
                if edge_value is ANY:
                    return None
                else:
                    assert_(edge_value is MATCH)
            else:
                path_fill_ins[END][k] = edge_value
        else:
            assert_(not isinstance(path_value, Placeholder))
            # Concrete values match placeholders, and identical concrete
            # values
            if (not isinstance(edge_value, Placeholder)
                and path_value != edge_value):
                return None
    # propagate any new constraints backwards through the path
    if path_fill_ins:
        new_nodes = []
        for node in path.nodes:
            new_node = dict(node)
            for placeholder, replacements in path_fill_ins.items():
                for key, value in replacements.items():
                    if new_node.get(key) is placeholder:
                        new_node[key] = value
            new_nodes.append(new_node)
        path = Path(new_nodes, path.transforms)
    # now extend the path forward by adding a new node
    new_path_end = dict(edge.end)
    # Propagate forward MATCH values
    _fill_values_from(new_path_end, MATCH, path.nodes[-1])
    # Add any new END values
    _replace_values(new_path_end, {ANY: END})
    return Path(path.nodes + [new_path_end],
                path.transforms + [edge.transform])

def test_try_extend_path():
    assert_(try_extend_path(
        Path([{"name": "s", "a": 1}], []),
        Edge({"name": "s", "a": 1}, {"name": "e", "b": 2}, "t1"))
            == Path([{"name": "s", "a": 1}, {"name": "e", "b": 2}],
                    ["t1"]))

    # "a" mismatch
    assert_(try_extend_path(
        Path([{"name": "s", "a": 1}], []),
        Edge({"name": "s", "a": 2}, {"name": "e", "b": 2}, "t1"))
            is None)

    # keys mismatch
    assert_(try_extend_path(
        Path([{"name": "s", "a": 1}], []),
        Edge({"name": "s", "b": 2}, {"name": "e", "b": 2}, "t1"))
            is None)

    # match ANY
    assert_(try_extend_path(
        Path([{"name": "s", "a": 1}], []),
        Edge({"name": "s", "a": ANY}, {"name": "e", "b": 2}, "t1"))
            == Path([{"name": "s", "a": 1}, {"name": "e", "b": 2}],
                    ["t1"]))

    # ANY -> END
    assert_(try_extend_path(
        Path([{"name": "s", "a": 1}], []),
        Edge({"name": "s", "a": 1}, {"name": "e", "b": ANY}, "t1"))
            == Path([{"name": "s", "a": 1}, {"name": "e", "b": END}],
                    ["t1"]))

    # propagate value through MATCH
    assert_(try_extend_path(
        Path([{"name": "s", "a": 1}], []),
        Edge({"name": "s", "a": MATCH}, {"name": "e", "a": MATCH}, "t1"))
            == Path([{"name": "s", "a": 1}, {"name": "e", "a": 1}],
                    ["t1"]))

    # propagate placeholder through MATCH
    assert_(try_extend_path(
        Path([{"name": "s", "a": START}], []),
        Edge({"name": "s", "a": MATCH}, {"name": "e", "a": MATCH}, "t1"))
            == Path([{"name": "s", "a": START}, {"name": "e", "a": START}],
                    ["t1"]))

    assert_(try_extend_path(
        Path([{"name": "s", "a": END}], []),
        Edge({"name": "s", "a": MATCH}, {"name": "e", "a": MATCH}, "t1"))
            == Path([{"name": "s", "a": END}, {"name": "e", "a": END}],
                    ["t1"]))

    # losing an END is no good
    assert_(try_extend_path(
        Path([{"name": "s", "a": END}], []),
        Edge({"name": "s", "a": ANY}, {"name": "e", "b": 1}, "t1"))
            is None)

    # replacing an END or START with a concrete value works
    assert_(try_extend_path(
        Path([{"name": "s", "a": START, "b": START}], []),
        Edge({"name": "s", "a": 1, "b": ANY}, {"name": "e", "c": 1}, "t1"))
            == Path([{"name": "s", "a": 1, "b": START},
                     {"name": "e", "c": 1}],
                    ["t1"]))

    assert_(try_extend_path(
        Path([{"name": "s", "a": END, "b": END}], []),
        Edge({"name": "s", "a": 1, "b": 2}, {"name": "e", "c": 1}, "t1"))
            == Path([{"name": "s", "a": 1, "b": 2},
                     {"name": "e", "c": 1}],
                    ["t1"]))

    # and this reaches backwards through the path
    assert_(try_extend_path(
        Path([{"name": "1", "a": START, "b": START},
              {"name": "2", "a": START},
              {"name": "3", "a": START, "b": END}],
             ["t1", "t2"]),
        Edge({"name": "3", "a": 1, "b": 2}, {"name": "4", "c": 1}, "t3"))
            == Path([{"name": "1", "a": 1, "b": START},
                     {"name": "2", "a": 1},
                     {"name": "3", "a": 1, "b": 2},
                     {"name": "4", "c": 1}],
                    ["t1", "t2", "t3"]))


def pairwise_shortest_paths(edges):
    # start node name -> [Edge]
    edges_from = defaultdict(list)

    # (frozen start node, frozen end node) -> Path
    shortest_paths = {}
    # list of Path objects
    todo_next = []

    # Simple implementation of breadth-first-search:
    def observe_path(path):
        check_path(path)
        #print("  new path: %r" % (path,))
        key = (tuple(path.nodes[0].items()), tuple(path.nodes[-1].items()))
        if key not in shortest_paths:
            #print("    ...accepted")
            shortest_paths[key] = path
            todo_next.append(path)
        else:
            #print("    ...rejected")
            pass

    for edge in edges:
        edges_from[edge.start["name"]].append(edge)
        observe_path(trivial_path(edge.start))

    while todo_next:
        todo_now = todo_next
        todo_next = []
        for path in todo_now:
            #print("extending: %s" % (path,))
            for edge in edges_from[path.nodes[-1]["name"]]:
                new_path = try_extend_path(path, edge)
                #print("  + %s -> %s" % (edge, new_path))
                if new_path is not None:
                    observe_path(new_path)


    shortest_paths_by_name = {}
    for path in shortest_paths.values():
        key = (path.nodes[0]["name"], path.nodes[-1]["name"])
        shortest_paths_by_name.setdefault(key, []).append(path)

    return shortest_paths_by_name

def test_pairwise_shortest_paths():
    # a -> b, b -> c
    assert_(pairwise_shortest_paths([
        Edge({"name": "a"}, {"name": "b"}, "ab"),
        Edge({"name": "b"}, {"name": "c"}, "bc"),
    ]) == {
        ("a", "a"): [Path([{"name": "a"}], [])],
        ("b", "b"): [Path([{"name": "b"}], [])],
        ("a", "b"): [Path([{"name": "a"}, {"name": "b"}], ["ab"])],
        ("b", "c"): [Path([{"name": "b"}, {"name": "c"}], ["bc"])],
        ("a", "c"): [Path([{"name": "a"}, {"name": "b"}, {"name": "c"}],
                          ["ab", "bc"])],
        })

    # a -> b, b -> c, a -> c
    # finds shortest a -> c path
    assert_(pairwise_shortest_paths([
        Edge({"name": "a"}, {"name": "b"}, "ab"),
        Edge({"name": "b"}, {"name": "c"}, "bc"),
        Edge({"name": "a"}, {"name": "c"}, "ac"),
    ]) == {
        ("a", "a"): [Path([{"name": "a"}], [])],
        ("b", "b"): [Path([{"name": "b"}], [])],
        ("a", "b"): [Path([{"name": "a"}, {"name": "b"}], ["ab"])],
        ("b", "c"): [Path([{"name": "b"}, {"name": "c"}], ["bc"])],
        ("a", "c"): [Path([{"name": "a"}, {"name": "c"}], ["ac"])],
        })

    # MATCH propagation
    assert_(pairwise_shortest_paths([
        Edge({"name": "a", "x": MATCH}, {"name": "b", "x": MATCH}, "ab"),
        Edge({"name": "b", "x": ANY}, {"name": "c"}, "bc"),
    ]) == {
        ("a", "a"): [Path([{"name": "a", "x": START}], [])],
        ("b", "b"): [Path([{"name": "b", "x": START}], [])],
        ("a", "b"): [Path([{"name": "a", "x": START},
                           {"name": "b", "x": START}],
                          ["ab"])],
        ("b", "c"): [Path([{"name": "b", "x": START},
                           {"name": "c"}],
                          ["bc"])],
        ("a", "c"): [Path([{"name": "a", "x": START},
                           {"name": "b", "x": START},
                           {"name": "c"}],
                          ["ab", "bc"])],
        })

    # a -> b, b -> c, a -> c
    # but a->c direct edge requires specific setting for x, so we get two
    # different paths
    paths = pairwise_shortest_paths([
        Edge({"name": "a", "x": ANY}, {"name": "b"}, "ab"),
        Edge({"name": "b"}, {"name": "c", "x": ANY}, "bc"),
        Edge({"name": "a", "x": 1}, {"name": "c", "x": 1}, "ac"),
    ])
    assert_(len(paths[("a", "c")]) == 3)
    assert_(Path([{"name": "a", "x": START},
                  {"name": "b"},
                  {"name": "c", "x": END}],
                 ["ab", "bc"])
            in paths[("a", "c")])
    assert_(Path([{"name": "a", "x": 1},
                  {"name": "c", "x": 1}],
                 ["ac"])
            in paths[("a", "c")])
    # The third one is a redundant a|x=1 -> b -> c|x=END


################################################################
# Paths + concrete nodes
################################################################

def concretize_path_node(path_node, start_concrete_node, end_concrete_node):
    concrete = dict(path_node)
    _fill_values_from(concrete, START, start_concrete_node)
    _fill_values_from(concrete, END, end_concrete_node)
    return concrete

def test_concretize_path_node():
    assert_(concretize_path_node(
        {"name": "foo",   "a": START, "b": END,  "c": "given"},
        {"name": "start", "a": "sa",  "b": "sb", "c": "sc"},
        {"name": "end",   "a": "ea",  "b": "eb", "c": "ec"})
            == {"name": "foo", "a": "sa", "b": "eb", "c": "given"})

def safe_equal(a, b):
    if isinstance(a, np.ndarray) or isinstance(b, np.ndarray):
        return np.array_equal(a, b)
    else:
        return a == b

def transform_kwargs(start_concrete_node, end_concrete_node):
    kwargs = {}
    for k, v in start_concrete_node.items():
        if k != "name":
            kwargs[k] = v
    for k, v in end_concrete_node.items():
        if k != "name":
            if k in kwargs:
                assert_(safe_equal(kwargs[k], v))
            else:
                kwargs[k] = v
    return kwargs

def test_transform_kwargs():
    from nose.tools import assert_raises

    assert_(transform_kwargs({"name": "start"}, {"name": "end"}) == {})
    assert_(transform_kwargs(
        {"name": "start", "a": 1, "b": 2},
        {"name": "end", "b": 2, "c": 3}
        ) == {"a": 1, "b": 2, "c": 3})
    assert_raises(AssertionError, transform_kwargs,
                  {"name": "start", "a": 1},
                  {"name": "end",   "a": 2})

def path_matches(path, desired_concrete_start, desired_concrete_end):
    # check that path[0] and path[-1] match concrete_start and concrete_end
    # and also that this holds even after filling in the values
    path_concrete_start = concretize_path_node(path.nodes[0],
                                               desired_concrete_start,
                                               desired_concrete_end)
    path_concrete_end = concretize_path_node(path.nodes[-1],
                                             desired_concrete_start,
                                             desired_concrete_end)
    return ((path_concrete_start == desired_concrete_start)
            and (path_concrete_end == desired_concrete_end))

def test_path_matches():
    assert_(path_matches(Path(({"name": "start"}, {"name": "end"}), ("t1",)),
                         {"name": "start"}, {"name": "end"}))
    assert_(path_matches(
        Path(({"name": "start", "a": START},
              {"name": "end", "b": END}),
             ("t1",)),
        {"name": "start", "a": 1},
        {"name": "end", "b": 2}))

    assert_(path_matches(
        Path(({"name": "start", "a": START},
              {"name": "end", "a": START}),
             ("t1",)),
        {"name": "start", "a": 1},
        {"name": "end", "a": 1}))

    assert_(not path_matches(
        Path(({"name": "start", "a": START},
              {"name": "end", "b": END}),
             ("t1",)),
        {"name": "start_mismatch", "a": 1},
        {"name": "end", "b": 2}))

    assert_(not path_matches(
        Path(({"name": "start", "a": 1},
              {"name": "end", "b": 2}),
             ("t1",)),
        {"name": "start", "a": 1},
        {"name": "end", "b": 22}))

    assert_(not path_matches(
        Path(({"name": "start", "a": START},
              {"name": "end", "a": START}),
             ("t1",)),
        {"name": "start", "a": 1},
        {"name": "end", "a": 2}))

class Transform(object):
    def __init__(self, nodes, transforms, kwargses):
        self.nodes = nodes
        self._transforms = transforms
        self._kwargses = kwargses

    def __call__(self, x):
        for transform, kwargs in zip(self._transforms, self._kwargses):
            x = transform(x, **kwargs)
        return x

def test_Transform():
    log = []
    def t1(x, **kwargs):
        log.append(("t1", x, kwargs))
        return x * 2
    def t2(x, **kwargs):
        log.append(("t2", x, kwargs))
        return x * 2

    t = Transform(["a", "b", "c"],
                  [t1, t2],
                  [{"t1_arg": 1}, {"t2_arg": 2}])
    assert t("x") == "xxxx"
    assert log == [
        ("t1", "x", {"t1_arg": 1}),
        ("t2", "xx", {"t2_arg": 2}),
        ]

################################################################
# Top-level object
################################################################

class TransformGraph(object):
    def __init__(self, edges, rank_constraints=[]):
        # This is what's used for actual calculations
        self._shortest_paths = pairwise_shortest_paths(edges)

        # All the rest is used only for the dot output
        self._rank_constraints = rank_constraints
        self._edges = edges
        self._nodes = []
        seen_names = set()
        for edge in edges:
            check_edge(edge)
            for node in (edge.start, edge.end):
                if node["name"] not in seen_names:
                    seen_names.add(node["name"])
                    self._nodes.append(node)

    def get_transform(self, start, end):
        check_concrete_node(start)
        check_concrete_node(end)
        if start == end:
            return Transform([start], [], [])

        start_name = start["name"]
        end_name = end["name"]
        best_path = None
        for path in self._shortest_paths.get((start_name, end_name), []):
            if path_matches(path, start, end):
                if best_path is None or len(best_path.nodes) > len(path.nodes):
                    best_path = path
        if best_path is None:
            raise ValueError("No path found from %r -> %r" % (start, end))
        concrete_nodes = []
        for path_node in best_path.nodes:
            concrete_nodes.append(concretize_path_node(path_node, start, end))
        kwargses = []
        for i in range(len(concrete_nodes) - 1):
            kwargs = transform_kwargs(concrete_nodes[i],
                                      concrete_nodes[i + 1])
            kwargses.append(kwargs)
        return Transform(concrete_nodes, best_path.transforms, kwargses)

    def dump_dot(self, f): # pragma: no cover
        f.write("digraph {\n")
        for node in self._nodes:
            # Laziness: assumes names don't need more quoting
            attr_names = set(node)
            attr_names.remove("name")
            html = "<b>%s</b>" % (node["name"],)
            for attr_name in sorted(attr_names):
                html += "<br/>  <i>%s</i>" % (attr_name,)
            f.write("  \"%s\" [ label=<%s> ]\n" % (node["name"], html))
        for edge in self._edges:
            # FIXME: label edges with attribute information
            # taillabel=<...>, headlabel=<...>, ...
            f.write("  \"%s\" -> \"%s\"\n"
                    % (edge.start["name"], edge.end["name"]))
        for rank_constraint in self._rank_constraints:
            f.write("  {rank=same; "
                    + ", ".join(["\"%s\"" % (c,) for c in rank_constraint])
                    + "}\n")
        f.write("}\n")
